import logging

from twisted.internet import defer, ssl
from twisted.internet.protocol import Protocol, ReconnectingClientFactory

from apns.errorresponse import ErrorResponse
from apns.listenable import Listenable


logger = logging.getLogger(__name__)


class GatewayClientError(Exception):
    """To be thrown upon failures on communication with APN gateway."""
    pass


class GatewayClientNotSetError(GatewayClientError):
    """
    Thrown when attempted to send a notification while connection is not
    established.
    """
    pass


class GatewayClient(Protocol):
    """
    Implements client-side of APN gateway protocol. Should be spawned by
    GatewayClientFactory and generally should not be used standalone.
    """

    @defer.inlineCallbacks
    def connectionMade(self):
        logger.debug('Gateway connection made: %s:%d', self.factory.hostname,
                     self.factory.port)
        yield self.factory.connectionMade(self)

    def send(self, notification):
        stream = notification.to_binary_string()
        self.transport.write(stream)

    @defer.inlineCallbacks
    def dataReceived(self, data):
        error = ErrorResponse()
        error.from_binary_string(data)
        yield self.factory.errorReceived(error)


class GatewayClientFactory(ReconnectingClientFactory, Listenable):
    """Allows connecting to the APN gateway and sending notifications."""
    protocol = GatewayClient
    maxDelay = 10
    ENDPOINTS = {
        'pub': ('gateway.push.apple.com', 2195),
        'dev': ('gateway.sandbox.push.apple.com', 2195)
    }
    EVENT_ERROR_RECEIVED = 'error received'
    EVENT_CONNECTION_MADE = 'connection made'
    EVENT_CONNECTION_LOST = 'connection lost'

    def __init__(self, endpoint, pem):
        """
        Init an instance of GatewayClientFactory.
        :param endpoint: Either 'pub' for production or 'dev' for development.
        :param pem: Path to a provider private certificate file.
        """
        Listenable.__init__(self)
        self.hostname, self.port = self.ENDPOINTS[endpoint]
        self.client = None

        with open(pem) as f:
            self.certificate = ssl.PrivateCertificate.loadPEM(f.read())

    @defer.inlineCallbacks
    def connectionMade(self, client):
        self.client = client
        yield self.dispatchEvent(self.EVENT_CONNECTION_MADE)

    @defer.inlineCallbacks
    def _onConnectionLost(self):
        self.client = None
        yield self.dispatchEvent(self.EVENT_CONNECTION_LOST)

    @defer.inlineCallbacks
    def errorReceived(self, error):
        logger.debug('Gateway error received: %s', error)
        yield self.dispatchEvent(self.EVENT_ERROR_RECEIVED, error)

    @defer.inlineCallbacks
    def clientConnectionFailed(self, connector, reason):
        logger.debug('Gateway connection failed: %s',
                     reason.getErrorMessage())
        yield self._onConnectionLost()
        yield ReconnectingClientFactory.clientConnectionFailed(self,
                                                               connector,
                                                               reason)
    @defer.inlineCallbacks
    def clientConnectionLost(self, connector, reason):
        logger.debug('Gateway connection lost: %s',
                     reason.getErrorMessage())
        yield self._onConnectionLost()
        yield ReconnectingClientFactory.clientConnectionLost(self,
                                                             connector,
                                                             reason)

    @property
    def connected(self):
        """Return True if connection with APN is established."""
        return self.client is not None

    def send(self, notification):
        """Send prepared notification to the APN."""
        logger.debug('Gateway send notification')

        if self.client is None:
            raise GatewayClientNotSetError()

        self.client.send(notification)



import logging

from twisted.internet import defer, ssl
from twisted.internet.protocol import Protocol, ReconnectingClientFactory

from apns.feedback import Feedback
from apns.listenable import Listenable


logger = logging.getLogger(__name__)


class FeedbackClient(Protocol):
    """
    Implements client-side of APN feedback service protocol. Should be spawned
    by FeedbackClientFactory and generally should not be used standalone.
    """
    def connectionMade(self):
        logger.debug('Feedback connection made: %s:%d', self.factory.hostname,
                     self.factory.port)

    @defer.inlineCallbacks
    def dataReceived(self, data):
        feedbacks = Feedback.from_binary_string(data)
        yield self.factory.feedbacksReceived(feedbacks)


class FeedbackClientFactory(ReconnectingClientFactory, Listenable):
    """
    Allows connecting to the APN feedback service and receiving feedback
    information. To process received feedbacks in your code, add a callback to
    EVENT_FEEDBACKS_RECEIVED.
    """
    protocol = FeedbackClient
    maxDelay = 600
    ENDPOINTS = {
        'pub': ('feedback.push.apple.com', 2196),
        'dev': ('feedback.sandbox.push.apple.com', 2196)
    }
    EVENT_FEEDBACKS_RECEIVED = 'feedbacks received'

    def __init__(self, endpoint, pem):
        """
        Init an instance of FeedbackClientFactory.
        :param endpoint: Either 'pub' for production or 'dev' for development.
        :param pem: Path to a provider private certificate file.
        """
        Listenable.__init__(self)
        self.hostname, self.port = self.ENDPOINTS[endpoint]
        self.client = None

        with open(pem) as f:
            self.certificate = ssl.PrivateCertificate.loadPEM(f.read())

    @defer.inlineCallbacks
    def feedbacksReceived(self, feedbacks):
        logger.debug('Feedbacks received: %s', feedbacks)
        yield self.dispatchEvent(self.EVENT_FEEDBACKS_RECEIVED, feedbacks)

    def clientConnectionFailed(self, connector, reason):
        logger.debug('Feedback connection failed: %s',
                     reason.getErrorMessage())
        return ReconnectingClientFactory.clientConnectionFailed(self,
                                                                connector,
                                                                reason)

    def clientConnectionLost(self, connector, reason):
        logger.debug('Feedback connection lost: %s',
                     reason.getErrorMessage())
        return ReconnectingClientFactory.clientConnectionLost(self,
                                                              connector,
                                                              reason)



import struct

from apns.commands import ERROR_RESPONSE


class ErrorResponseError(Exception):
    """To be thrown upon failures on error response processing."""
    pass


class ErrorResponseInvalidCommandError(ErrorResponseError):
    """
    Thrown while unpacking an error response, if the command field contains
    invalid value.
    """
    pass


class ErrorResponseInvalidCodeError(ErrorResponseError):
    """
    Thrown while unpacking an error response, if the status code field contains
    invalid value.
    """
    pass


class ErrorResponse(object):
    """
    A representation of the structure of an error response, as defined in the
    iOS documentation.
    """
    CODE_OK = 0
    CODE_PROCESSING_ERROR = 1
    CODE_MISSING_TOKEN = 2
    CODE_MISSING_TOPIC = 3
    CODE_MISSING_PAYLOAD = 4
    CODE_INVALID_TOKEN_SIZE = 5
    CODE_INVALID_TOPIC_SIZE = 6
    CODE_INVALID_PAYLOAD_SIZE = 7
    CODE_INVALID_TOKEN = 8
    CODE_SHUTDOWN = 10
    CODE_UNKNOWN = 255

    CODES = {
        CODE_OK: 'No errors encountered',
        CODE_PROCESSING_ERROR: 'Processing error',
        CODE_MISSING_TOKEN: 'Missing token',
        CODE_MISSING_TOPIC: 'Missing topic',
        CODE_MISSING_PAYLOAD: 'Missing payload',
        CODE_INVALID_TOKEN_SIZE: 'Invalid token size',
        CODE_INVALID_TOPIC_SIZE: 'Invalid topic size',
        CODE_INVALID_PAYLOAD_SIZE: 'Invalid payload size',
        CODE_INVALID_TOKEN: 'Invalid token',
        CODE_SHUTDOWN: 'Shutdown',
        CODE_UNKNOWN: 'Unknown'
    }

    FORMAT = '>BBI'
    COMMAND = ERROR_RESPONSE

    def __init__(self):
        self.code = self.name = self.identifier = None

    def __str__(self):
        return '<ErrorResponse: %s>' % self.name

    def from_binary_string(self, stream):
        """Unpack the error response from a stream."""
        command, code, identifier = struct.unpack(self.FORMAT, stream)

        if command != self.COMMAND:
            raise ErrorResponseInvalidCommandError()

        if code not in self.CODES:
            raise ErrorResponseInvalidCodeError()

        self.code = code
        self.name = self.CODES[code]
        self.identifier = identifier

    def to_binary_string(self, code, identifier):
        """Pack the error response to binary string and return it."""
        return struct.pack(self.FORMAT, self.COMMAND, code, identifier)



from collections import defaultdict

from twisted.internet import defer


class Listenable(object):
    """Implements basic listener/observer model for derivative classes."""

    def __init__(self):
        self.listeners = defaultdict(list)

    def listen(self, event, callback):
        """
        Assign a callback to an event.
        :param event: an event which triggers execution of the callback.
        Particular values are defined by derivative classes.
        :param callback: a callback to be fired when event occurs.
        """
        self.listeners[event].append(callback)

    def unlisten(self, event, callback):
        """
        Remove previously assigned callback.
        :return True in case the callback was successfully removed, False
        otherwise.
        """
        try:
            self.listeners[event].remove(callback)
        except ValueError:
            return False
        else:
            return True

    @defer.inlineCallbacks
    def dispatchEvent(self, event, *args):
        """
        Fire all callbacks assigned to a particular event. To be called by
        derivative classes.
        :param *args: Additional arguments to be passed to the callback
        function.
        """
        for callback in self.listeners[event]:
            yield callback(event, self, *args)



# -*- coding: utf-8 -*-
from distutils.core import setup


VERSION = '0.12'
URL = 'https://github.com/operasoftware/twisted-apns'
DOWNLOAD_URL = URL + '/tarball/' + VERSION


setup(
  name = 'twisted-apns',
  packages = ['apns'],
  version = VERSION,
  description = 'Twisted client for Apple Push Notification Service (APNs)',
  author = 'Michał Łowicki',
  author_email = 'mlowicki@opera.com',
  url = URL,
  download_url = DOWNLOAD_URL,
  keywords = ['twisted', 'apns'],
  classifiers = [],
  install_requires=['Twisted', 'pyOpenSSL']
)



from datetime import datetime
import binascii
import json
import struct

from apns.commands import NOTIFICATION
from apns.utils import datetime_to_timestamp


class NotificationError(Exception):
    """To be thrown upon failures on notification processing."""
    pass


class NotificationInvalidPriorityError(NotificationError):
    """
    Thrown while packing a notification, if the notification priority field is
    invalid.
    """
    pass


class NotificationPayloadNotSerializableError(NotificationError):
    """
    Thrown while packing a notification, if the notification payload field
    could not be serialized to JSON.
    """
    pass


class NotificationTokenUnhexlifyError(NotificationError):
    """
    Thrown while packing a notification, if the notification token field could
    not be converted to binary from its hex representation.
    """
    def __init__(self, msg):
        super(NotificationTokenUnhexlifyError, self).__init__(msg)


class NotificationInvalidCommandError(NotificationError):
    """
    Thrown while unpacking a notification, if the notification command field
    contains invalid value.
    """
    pass


class NotificationInvalidIdError(NotificationError):
    """
    Thrown while unpacking a notification, if the notification structure is
    invalid.
    """
    pass


class Notification(object):
    """
    A representation of the structure of a notification request, as defined in
    the iOS documentation.
    """
    COMMAND = NOTIFICATION
    PRIORITY_NORMAL = 5
    PRIORITY_IMMEDIATELY = 10

    PRIORITIES = (PRIORITY_NORMAL, PRIORITY_IMMEDIATELY)

    PAYLOAD = 2
    TOKEN = 1
    PRIORITY = 5
    NOTIFICATION_ID = 3
    EXPIRE = 4

    EXPIRE_IMMEDIATELY = 0

    def __init__(self, payload=None, token=None, expire=None,
                 priority=PRIORITY_NORMAL, iden=0):
        """
        Init an instance of Notification.
        :param payload: object containing structure of payload to be sent to
        remote device.
        :param token: string containing target device token in hex
        :param expire: notification expire time as UNIX timestamp, 0 means that
        notification expires immediately.
        :param priority: notification priority, as described in iOS
        documentation
        :param iden: notification ID, as described in iOS documentation
        """
        self.payload = payload
        self.token = token
        self.expire = expire
        self.priority = priority
        self.iden = iden

    def __str__(self):
        return '<Notification: %s>' % self.token

    def to_binary_string(self):
        """Pack the notification to binary form and return it as string."""
        if self.priority not in self.PRIORITIES:
            raise NotificationInvalidPriorityError()

        try:
            token = binascii.unhexlify(self.token)
        except TypeError as error:
            raise NotificationTokenUnhexlifyError(error)

        try:
            payload = json.dumps(self.payload)
        except TypeError:
            raise NotificationPayloadNotSerializableError()

        fmt = ">BIBH{0}sBH{1}sBHIBHIBHB".format(len(token), len(payload))

        expire = (0 if self.expire == self.EXPIRE_IMMEDIATELY else
                  datetime_to_timestamp(self.expire))

        # |COMMAND|FRAME-LEN|{token}|{payload}|{id:4}|{expire:4}|{priority:1}
        # 5 items, each 3 bytes prefix, then each item length
        length = 3*5 + len(token) + len(payload) + 4 + 4 + 1
        message = struct.pack(fmt, self.COMMAND, length,
                              self.TOKEN, len(token), token,
                              self.PAYLOAD, len(payload), payload,
                              self.NOTIFICATION_ID, 4, self.iden,
                              self.EXPIRE, 4, expire,
                              self.PRIORITY, 1, self.priority)
        return message

    def from_binary_string(self, notification):
        """Unpack the notification from binary string."""
        command = struct.unpack('>B', notification[0])[0]

        if command != self.COMMAND:
            raise NotificationInvalidCommandError()

        length = struct.unpack('>I', notification[1:5])[0]
        notification = notification[5:]
        offset = 0

        def next_item(offset):
            iden, length = struct.unpack('>BH', notification[offset:offset+3])
            offset += 3
            payload = notification[offset:offset+length]
            offset += length

            if iden == self.PAYLOAD:
                payload = struct.unpack('>{0}s'.format(length), payload)[0]
                self.payload = json.loads(payload)
            elif iden == self.TOKEN:
                payload = struct.unpack('>{0}s'.format(length), payload)[0]
                self.token = binascii.hexlify(payload)
            elif iden == self.PRIORITY:
                self.priority = struct.unpack('>B', payload)[0]
            elif iden == self.NOTIFICATION_ID:
                self.iden = struct.unpack('>I', payload)[0]
            elif iden == self.EXPIRE:
                payload = struct.unpack('>I', payload)[0]
                self.expire = (self.EXPIRE_IMMEDIATELY if payload == 0 else
                               datetime.fromtimestamp(payload))
            else:
                raise NotificationInvalidIdError()

            return offset

        while offset < length:
            offset = next_item(offset)
